"""
   An implementation of high-level API to SNMP v.1 message PDU objects
   (RFC1157)

   Copyright 1999-2003 by Ilya Etingof <ilya@glas.net>. See LICENSE for
   details.
"""
# Module public names
__all__ = [ 'GetRequestPduMixIn', 'GetNextRequestPduMixIn',
            'SetRequestPduMixIn', 'GetResponsePduMixIn',
            'ResponsePduMixIn', 'TrapPduMixIn', 'registerMixIns' ]

from pysnmp.proto import rfc1155, rfc1157
from pysnmp.proto.api import error

class RequestPduMixIn:
    def apiGenGetRequestId(self): return self['request_id'].get()
    apiGetRequestId = apiGenGetRequestId
    def apiGenSetRequestId(self, value): self['request_id'].set(value)
    apiSetRequestId = apiGenSetRequestId
    def apiGenGetVarBind(self):
        outVars = []
        for varBind in self['variable_bindings']:
            outVars.append((varBind['name'].get(),
                            varBind['value'].values()[0].values()[0]))
        return outVars
    
    apiGetVarBind = apiGenGetVarBind
    
    def apiGenSetVarBind(self, inVars):
        varBindList = rfc1157.VarBindList()
        for inVar in inVars:
            try:
                (name, value) = inVar
            except ValueError:
                raise error.BadArgumentError('A (name, value) style entry expected by %s' % (repr(inVar), self.__class__.__name__))

            # Default to empty payload
            if value is None:
                value = rfc1155.Null()

            # Handle IP address
            if isinstance(value, rfc1155.IpAddress):
                value = rfc1155.NetworkAddress(addr=value)
            
            if reduce(lambda x, y: x+y, map(lambda x, v=value: isinstance(v, x), rfc1155.SimpleSyntax.choiceComponents)):
                varBindList.append(rfc1157.VarBind(name=rfc1155.ObjectName(name), value=rfc1155.ObjectSyntax(simple=rfc1155.SimpleSyntax(t=value))))
            elif reduce(lambda x, y: x+y, map(lambda x, v=value: isinstance(v, x), rfc1155.ApplicationSyntax.choiceComponents)):
                varBindList.append(rfc1157.VarBind(name=rfc1155.ObjectName(name), value=rfc1155.ObjectSyntax(simple=rfc1155.ApplicationSyntax(t=value))))
            else:
                raise error.BadArgumentError('Unknown value type %s at %s' % (repr(value), self.__class__.__name__))
            
        self['variable_bindings'] = varBindList
    apiSetVarBind = apiGenSetVarBind
    
# Request PDU mix-ins
class GetRequestPduMixIn(RequestPduMixIn): pass
class GetNextRequestPduMixIn(RequestPduMixIn): pass
class SetRequestPduMixIn(RequestPduMixIn): pass

class GetResponsePduMixIn(RequestPduMixIn):
    def apiGenGetErrorStatus(self): return self['error_status'].get()
    apiGetErrorStatus = apiGenGetErrorStatus
    def apiGenSetErrorStatus(self, value): self['error_status'].set(value)
    apiSetErrorStatus = apiGenSetErrorStatus
    def apiGenGetErrorIndex(self): return self['error_index'].get()
    apiGetErrorIndex = apiGenGetErrorIndex
    def apiGenSetErrorIndex(self, value): self['error_index'].set(value)
    apiSetErrorIndex = apiGenSetErrorIndex

# A v2c-style alias
ResponsePduMixIn = GetResponsePduMixIn

class TrapPduMixIn(RequestPduMixIn):
    def apiGenGetEnterprise(self):
        return self['enterprise'].get()
    apiGetEnterprise = apiGenGetEnterprise
    def apiGenSetEnterprise(self, value):
        self['enterprise'].set(value)
    apiSetEnterprise = apiGenSetEnterprise
    def apiGenGetAgentAddr(self):
        return self['agent_addr']['internet'].get()
    apiGetAgentAddr = apiGenGetAgentAddr
    def apiGenSetAgentAddr(self, value):
        self['agent_addr']['internet'].set(value)
    apiSetAgentAddr = apiGenSetAgentAddr
    def apiGenGetGenericTrap(self):
        return self['generic_trap'].get()
    apiGetGenericTrap = apiGenGetGenericTrap
    def apiGenSetGenericTrap(self, value):
        self['generic_trap'].set(value)
    apiSetGenericTrap = apiGenSetGenericTrap
    def apiGenGetSpecificTrap(self):
        return self['specific_trap'].get()
    apiGetSpecificTrap = apiGenGetSpecificTrap
    def apiGenSetSpecificTrap(self, value):
        self['specific_trap'].set(value)
    apiSetSpecificTrap = apiGenSetSpecificTrap
    def apiGenGetTimeStamp(self):
        return self['time_stamp'].get()
    apiGetTimeStamp = apiGenGetTimeStamp
    def apiGenSetTimeStamp(self, value):
        self['time_stamp'].set(value)
    apiSetTimeStamp = apiGenSetTimeStamp

def mixIn():
    """Register this module's mix-in classes at their bases
    """
    mixInComps = [ (rfc1157.GetRequestPdu, GetRequestPduMixIn),
                   (rfc1157.GetNextRequestPdu, GetNextRequestPduMixIn),
                   (rfc1157.SetRequestPdu, SetRequestPduMixIn),
                   (rfc1157.GetResponsePdu, GetResponsePduMixIn),
                   (rfc1157.TrapPdu, TrapPduMixIn) ]

    for (baseClass, mixIn) in mixInComps:
        if mixIn not in baseClass.__bases__:
            baseClass.__bases__ = baseClass.__bases__ + (mixIn, )
