"""
   An implementation of high-level API to SNMP v.2c message PDU objects
   (RFC1905)

   Copyright 1999-2003 by Ilya Etingof <ilya@glas.net>. See LICENSE for
   details.
"""
# Module public names
__all__ = [ 'GetRequestPduMixIn', 'GetNextRequestPduMixIn',
            'SetRequestPduMixIn', 'ResponsePduMixIn',
            'GetBulkRequestPduMixIn', 'InformRequestPduMixIn',
            'ReportPduMixIn', 'SnmpV2TrapPduMixIn', 'registerMixIns' ]

from pysnmp.proto import rfc1902, rfc1905
from pysnmp.proto.api.generic import rfc1157
from pysnmp.proto.api import error

class RequestPduMixIn(rfc1157.RequestPduMixIn):
    def apiGenGetVarBind(self):
        outVars = []
        for varBind in self['variable_bindings']:
            # The value in v2c var may appear at different levels
            def getTerminalValue(value):
                while 1:
                    if isinstance(value, rfc1902.Choice):
                        value = value.values()[0]
                    else:
                        return value
                
            outVars.append((varBind['name'].get(),
                            getTerminalValue(varBind['value'])))

        return outVars

    apiGetVarBind = apiGenGetVarBind
    
    def apiGenSetVarBind(self, inVars):
        varBindList = rfc1905.VarBindList()
        for inVar in inVars:
            try:
                (name, value) = inVar
            except ValueError:
                raise error.BadArgumentError('A (name, value) style entry expected by %s' % (repr(inVar), self.__class__.__name__))

            # Default to empty payload
            if value is None: value = rfc1902.Null()

            if reduce(lambda x, y: x+y, map(lambda x, v=value: isinstance(v, x), rfc1905.BindValue.choiceComponents)):
                varBindList.append(rfc1905.VarBind(name=rfc1902.ObjectName(name), value=rfc1905.BindValue(t=value)))
            elif reduce(lambda x, y: x+y, map(lambda x, v=value: isinstance(v, x), rfc1902.SimpleSyntax.choiceComponents)):
                varBindList.append(rfc1905.VarBind(name=rfc1902.ObjectName(name), value=rfc1905.BindValue(syntax=rfc1902.ObjectSyntax(simple=rfc1902.SimpleSyntax(t=value)))))
            elif reduce(lambda x, y: x+y, map(lambda x, v=value: isinstance(v, x), rfc1902.ApplicationSyntax.choiceComponents)):
                varBindList.append(rfc1905.VarBind(name=rfc1902.ObjectName(name), value=rfc1905.BindValue(syntax=rfc1902.ObjectSyntax(app=rfc1902.ApplicationSyntax(t=value)))))
            else:
                raise error.BadArgumentError('Unknown value type %s at %s' % (repr(value), self.__class__.__name__))

        self['variable_bindings'] = varBindList

    apiSetVarBind = apiGenSetVarBind
        
# Request PDU mix-ins
class GetRequestPduMixIn(RequestPduMixIn): pass
class GetNextRequestPduMixIn(RequestPduMixIn): pass
class SetRequestPduMixIn(RequestPduMixIn): pass
class InformRequestPduMixIn(RequestPduMixIn): pass
class ReportPduMixIn(RequestPduMixIn): pass
class SnmpV2TrapPduMixIn(RequestPduMixIn): pass

class ResponsePduMixIn(RequestPduMixIn):
    def apiGenGetErrorStatus(self): return self['error_status'].get()
    apiGetErrorStatus = apiGenGetErrorStatus
    def apiGenSetErrorStatus(self, value): self['error_status'].set(value)
    apiSetErrorStatus = apiGenSetErrorStatus
    def apiGenGetErrorIndex(self): return self['error_index'].get()
    apiGetErrorIndex = apiGenGetErrorIndex
    def apiGenSetErrorIndex(self, value): self['error_index'].set(value)
    apiSetErrorIndex = apiGenSetErrorIndex

# A v1-style alias
GetResponsePduMixIn = ResponsePduMixIn
    
class GetBulkRequestPduMixIn(RequestPduMixIn):
    def apiGenGetNonRepeaters(self): return self['non_repeaters'].get()
    apiGetNonRepeaters = apiGenGetNonRepeaters
    def apiGenSetNonRepeaters(self, value): self['non_repeaters'].set(value)
    apiSetNonRepeaters = apiGenSetNonRepeaters
    def apiGenGetMaxRepetitions(self): return self['max_repetitions'].get()
    apiGetMaxRepetitions = apiGenGetMaxRepetitions
    def apiGenSetMaxRepetitions(self, value):
        self['max_repetitions'].set(value)
    apiSetMaxRepetitions = apiGenSetMaxRepetitions

def mixIn():
    """Register this module's mix-in classes at their bases
    """
    mixInComps = [ (rfc1905.GetRequestPdu, GetRequestPduMixIn),
                   (rfc1905.GetNextRequestPdu, GetNextRequestPduMixIn),
                   (rfc1905.SetRequestPdu, SetRequestPduMixIn),
                   (rfc1905.ResponsePdu, ResponsePduMixIn),
                   (rfc1905.GetBulkRequestPdu, GetBulkRequestPduMixIn),
                   (rfc1905.InformRequestPdu, InformRequestPduMixIn),
                   (rfc1905.ReportPdu, ReportPduMixIn),
                   (rfc1905.SnmpV2TrapPdu, SnmpV2TrapPduMixIn) ]

    for (baseClass, mixIn) in mixInComps:
        if mixIn not in baseClass.__bases__:
            baseClass.__bases__ = baseClass.__bases__ + (mixIn, )
